@@ -5,13 +5,19 @@ __metaclass__ = type
 
 import socket
 import time
+import logging
+import traceback as traceback_
+from collections import namedtuple
 
 from six.moves import range, http_client, urllib
 
 import six
 import pytest
+from jaraco.text import trim, unwrap
 
 from cheroot.test import helper, webtest
+from cheroot._compat import IS_PYPY
+import cheroot.server
 
 
 timeout = 1
@@ -26,7 +32,7 @@ class Controller(helper.Controller):
         return 'Hello, world!'
 
     def pov(req, resp):
-        """Render pov value."""
+        """Render ``pov`` value."""
         return pov
 
     def stream(req, resp):
@@ -43,8 +49,10 @@ class Controller(helper.Controller):
     def upload(req, resp):
         """Process file upload and render thank."""
         if not req.environ['REQUEST_METHOD'] == 'POST':
-            raise AssertionError("'POST' != request.method %r" %
-                                 req.environ['REQUEST_METHOD'])
+            raise AssertionError(
+                "'POST' != request.method %r" %
+                req.environ['REQUEST_METHOD'],
+            )
         return "thanks for '%s'" % req.environ['wsgi.input'].read()
 
     def custom_204(req, resp):
@@ -82,9 +90,9 @@ class Controller(helper.Controller):
         WSGI 1.0 is a mess around unicode. Create endpoints
         that match the PATH_INFO that it produces.
         """
-        if six.PY3:
-            return string.encode('utf-8').decode('latin-1')
-        return string
+        if six.PY2:
+            return string
+        return string.encode('utf-8').decode('latin-1')
 
     handlers = {
         '/hello': hello,
@@ -103,9 +111,33 @@ class Controller(helper.Controller):
     }
 
 
+class ErrorLogMonitor:
+    """Mock class to access the server error_log calls made by the server."""
+
+    ErrorLogCall = namedtuple('ErrorLogCall', ['msg', 'level', 'traceback'])
+
+    def __init__(self):
+        """Initialize the server error log monitor/interceptor.
+
+        If you need to ignore a particular error message use the property
+        ``ignored_msgs` by appending to the list the expected error messages.
+        """
+        self.calls = []
+        # to be used the the teardown validation
+        self.ignored_msgs = []
+
+    def __call__(self, msg='', level=logging.INFO, traceback=False):
+        """Intercept the call to the server error_log method."""
+        if traceback:
+            tblines = traceback_.format_exc()
+        else:
+            tblines = ''
+        self.calls.append(ErrorLogMonitor.ErrorLogCall(msg, level, tblines))
+
+
 @pytest.fixture
-def testing_server(wsgi_server_client):
-    """Attach a WSGI app to the given server and pre-configure it."""
+def raw_testing_server(wsgi_server_client):
+    """Attach a WSGI app to the given server and preconfigure it."""
     app = Controller()
 
     def _timeout(req, resp):
@@ -116,9 +148,37 @@ def testing_server(wsgi_server_client):
     wsgi_server.max_request_body_size = 1001
     wsgi_server.timeout = timeout
     wsgi_server.server_client = wsgi_server_client
+    wsgi_server.keep_alive_conn_limit = 2
+
     return wsgi_server
 
 
+@pytest.fixture
+def testing_server(raw_testing_server, monkeypatch):
+    """Modify the "raw" base server to monitor the error_log messages.
+
+    If you need to ignore a particular error message use the property
+    ``testing_server.error_log.ignored_msgs`` by appending to the list
+    the expected error messages.
+    """
+    # patch the error_log calls of the server instance
+    monkeypatch.setattr(raw_testing_server, 'error_log', ErrorLogMonitor())
+
+    yield raw_testing_server
+
+    # Teardown verification, in case that the server logged an
+    # error that wasn't notified to the client or we just made a mistake.
+    for c in raw_testing_server.error_log.calls:
+        if c.level <= logging.WARNING:
+            continue
+
+        assert c.msg in raw_testing_server.error_log.ignored_msgs, (
+            'Found error in the error log: '
+            "message = '{c.msg}', level = '{c.level}'\n"
+            '{c.traceback}'.format(**locals()),
+        )
+
+
 @pytest.fixture
 def test_client(testing_server):
     """Get and return a test client out of the given server."""
@@ -132,8 +192,10 @@ def header_exists(header_name, headers):
 
 def header_has_value(header_name, header_value, headers):
     """Check that a header with a given value is present."""
-    return header_name.lower() in (k.lower() for (k, v) in headers
-                                   if v == header_value)
+    return header_name.lower() in (
+        k.lower() for (k, v) in headers
+        if v == header_value
+    )
 
 
 def test_HTTP11_persistent_connections(test_client):
@@ -145,7 +207,7 @@ def test_HTTP11_persistent_connections(test_client):
 
     # Make the first request and assert there's no "Connection: close".
     status_line, actual_headers, actual_resp_body = test_client.get(
-        '/pov', http_conn=http_connection
+        '/pov', http_conn=http_connection,
     )
     actual_status = int(status_line[:3])
     assert actual_status == 200
@@ -155,7 +217,7 @@ def test_HTTP11_persistent_connections(test_client):
 
     # Make another request on the same connection.
     status_line, actual_headers, actual_resp_body = test_client.get(
-        '/page1', http_conn=http_connection
+        '/page1', http_conn=http_connection,
     )
     actual_status = int(status_line[:3])
     assert actual_status == 200
@@ -166,7 +228,7 @@ def test_HTTP11_persistent_connections(test_client):
     # Test client-side close.
     status_line, actual_headers, actual_resp_body = test_client.get(
         '/page2', http_conn=http_connection,
-        headers=[('Connection', 'close')]
+        headers=[('Connection', 'close')],
     )
     actual_status = int(status_line[:3])
     assert actual_status == 200
@@ -184,7 +246,7 @@ def test_HTTP11_persistent_connections(test_client):
     (
         False,  # Without Content-Length
         True,  # With Content-Length
-    )
+    ),
 )
 def test_streaming_11(test_client, set_cl):
     """Test serving of streaming responses with HTTP/1.1 protocol."""
@@ -195,7 +257,7 @@ def test_streaming_11(test_client, set_cl):
 
     # Make the first request and assert there's no "Connection: close".
     status_line, actual_headers, actual_resp_body = test_client.get(
-        '/pov', http_conn=http_connection
+        '/pov', http_conn=http_connection,
     )
     actual_status = int(status_line[:3])
     assert actual_status == 200
@@ -208,7 +270,7 @@ def test_streaming_11(test_client, set_cl):
         # When a Content-Length is provided, the content should stream
         # without closing the connection.
         status_line, actual_headers, actual_resp_body = test_client.get(
-            '/stream?set_cl=Yes', http_conn=http_connection
+            '/stream?set_cl=Yes', http_conn=http_connection,
         )
         assert header_exists('Content-Length', actual_headers)
         assert not header_has_value('Connection', 'close', actual_headers)
@@ -222,7 +284,7 @@ def test_streaming_11(test_client, set_cl):
         # streamed output will either close the connection, or use
         # chunked encoding, to determine transfer-length.
         status_line, actual_headers, actual_resp_body = test_client.get(
-            '/stream', http_conn=http_connection
+            '/stream', http_conn=http_connection,
         )
         assert not header_exists('Content-Length', actual_headers)
         assert actual_status == 200
@@ -250,7 +312,7 @@ def test_streaming_11(test_client, set_cl):
         # TODO: figure out how can this be possible on an closed connection
         # (chunked_response case)
         status_line, actual_headers, actual_resp_body = test_client.head(
-            '/stream', http_conn=http_connection
+            '/stream', http_conn=http_connection,
         )
         assert actual_status == 200
         assert status_line[4:] == 'OK'
@@ -263,7 +325,7 @@ def test_streaming_11(test_client, set_cl):
     (
         False,  # Without Content-Length
         True,  # With Content-Length
-    )
+    ),
 )
 def test_streaming_10(test_client, set_cl):
     """Test serving of streaming responses with HTTP/1.0 protocol."""
@@ -335,8 +397,14 @@ def test_streaming_10(test_client, set_cl):
     'http_server_protocol',
     (
         'HTTP/1.0',
-        'HTTP/1.1',
-    )
+        pytest.param(
+            'HTTP/1.1',
+            marks=pytest.mark.xfail(
+                IS_PYPY and not six.PY2,
+                reason='Fails under PyPy for unknown reason',
+            ),
+        ),
+    ),
 )
 def test_keepalive(test_client, http_server_protocol):
     """Test Keep-Alive enabled connections."""
@@ -372,6 +440,11 @@ def test_keepalive(test_client, http_server_protocol):
     assert status_line[4:] == 'OK'
     assert actual_resp_body == pov.encode()
     assert header_has_value('Connection', 'Keep-Alive', actual_headers)
+    assert header_has_value(
+        'Keep-Alive',
+        'timeout={test_client.server_instance.timeout}'.format(**locals()),
+        actual_headers,
+    )
 
     # Remove the keep-alive header again.
     status_line, actual_headers, actual_resp_body = test_client.get(
@@ -383,21 +456,101 @@ def test_keepalive(test_client, http_server_protocol):
     assert status_line[4:] == 'OK'
     assert actual_resp_body == pov.encode()
     assert not header_exists('Connection', actual_headers)
+    assert not header_exists('Keep-Alive', actual_headers)
 
     test_client.server_instance.protocol = original_server_protocol
 
 
+def test_keepalive_conn_management(test_client):
+    """Test management of Keep-Alive connections."""
+    test_client.server_instance.timeout = 2
+
+    def connection():
+        # Initialize a persistent HTTP connection
+        http_connection = test_client.get_connection()
+        http_connection.auto_open = False
+        http_connection.connect()
+        return http_connection
+
+    def request(conn, keepalive=True):
+        status_line, actual_headers, actual_resp_body = test_client.get(
+            '/page3', headers=[('Connection', 'Keep-Alive')],
+            http_conn=conn, protocol='HTTP/1.0',
+        )
+        actual_status = int(status_line[:3])
+        assert actual_status == 200
+        assert status_line[4:] == 'OK'
+        assert actual_resp_body == pov.encode()
+        if keepalive:
+            assert header_has_value('Connection', 'Keep-Alive', actual_headers)
+            assert header_has_value(
+                'Keep-Alive',
+                'timeout={test_client.server_instance.timeout}'.
+                format(**locals()),
+                actual_headers,
+            )
+        else:
+            assert not header_exists('Connection', actual_headers)
+            assert not header_exists('Keep-Alive', actual_headers)
+
+    disconnect_errors = (
+        http_client.BadStatusLine,
+        http_client.CannotSendRequest,
+        http_client.NotConnected,
+    )
+
+    # Make a new connection.
+    c1 = connection()
+    request(c1)
+
+    # Make a second one.
+    c2 = connection()
+    request(c2)
+
+    # Reusing the first connection should still work.
+    request(c1)
+
+    # Creating a new connection should still work, but we should
+    # have run out of available connections to keep alive, so the
+    # server should tell us to close.
+    c3 = connection()
+    request(c3, keepalive=False)
+
+    # Show that the third connection was closed.
+    with pytest.raises(disconnect_errors):
+        request(c3)
+
+    # Wait for some of our timeout.
+    time.sleep(1.2)
+
+    # Refresh the second connection.
+    request(c2)
+
+    # Wait for the remainder of our timeout, plus one tick.
+    time.sleep(1.2)
+
+    # First connection should now be expired.
+    with pytest.raises(disconnect_errors):
+        request(c1)
+
+    # But the second one should still be valid.
+    request(c2)
+
+    # Restore original timeout.
+    test_client.server_instance.timeout = timeout
+
+
 @pytest.mark.parametrize(
     'timeout_before_headers',
     (
         True,
         False,
-    )
+    ),
 )
 def test_HTTP11_Timeout(test_client, timeout_before_headers):
     """Check timeout without sending any data.
 
-    The server will close the conn with a 408.
+    The server will close the connection with a 408.
     """
     conn = test_client.get_connection()
     conn.auto_open = False
@@ -516,7 +669,7 @@ def test_HTTP11_Timeout_after_request(test_client):
 def test_HTTP11_pipelining(test_client):
     """Test HTTP/1.1 pipelining.
 
-    httplib doesn't support this directly.
+    :py:mod:`http.client` doesn't support this directly.
     """
     conn = test_client.get_connection()
 
@@ -528,7 +681,7 @@ def test_HTTP11_pipelining(test_client):
     for trial in range(5):
         # Put next request
         conn._output(
-            ('GET /hello?%s HTTP/1.1' % trial).encode('iso-8859-1')
+            ('GET /hello?%s HTTP/1.1' % trial).encode('iso-8859-1'),
         )
         conn._output(('Host: %s' % conn.host).encode('ascii'))
         conn._send_output()
@@ -537,9 +690,9 @@ def test_HTTP11_pipelining(test_client):
         response = conn.response_class(conn.sock, method='GET')
         # there is a bug in python3 regarding the buffering of
         # ``conn.sock``. Until that bug get's fixed we will
-        # monkey patch the ``reponse`` instance.
+        # monkey patch the ``response`` instance.
         # https://bugs.python.org/issue23377
-        if six.PY3:
+        if not six.PY2:
             response.fp = conn.sock.makefile('rb', 0)
         response.begin()
         body = response.read(13)
@@ -561,7 +714,7 @@ def test_100_Continue(test_client):
     conn = test_client.get_connection()
 
     # Try a page without an Expect request header first.
-    # Note that httplib's response.begin automatically ignores
+    # Note that http.client's response.begin automatically ignores
     # 100 Continue responses, so we must manually check for it.
     conn.putrequest('POST', '/upload', skip_host=True)
     conn.putheader('Host', conn.host)
@@ -592,7 +745,8 @@ def test_100_Continue(test_client):
         if line:
             pytest.fail(
                 '100 Continue should not output any headers. Got %r' %
-                line)
+                line,
+            )
         else:
             break
 
@@ -615,7 +769,7 @@ def test_100_Continue(test_client):
     (
         0,
         1001,
-    )
+    ),
 )
 def test_readall_or_close(test_client, max_request_body_size):
     """Test a max_request_body_size of 0 (the default) and 1001."""
@@ -691,7 +845,7 @@ def test_No_Message_Body(test_client):
 
     # Make the first request and assert there's no "Connection: close".
     status_line, actual_headers, actual_resp_body = test_client.get(
-        '/pov', http_conn=http_connection
+        '/pov', http_conn=http_connection,
     )
     actual_status = int(status_line[:3])
     assert actual_status == 200
@@ -701,7 +855,7 @@ def test_No_Message_Body(test_client):
 
     # Make a 204 request on the same connection.
     status_line, actual_headers, actual_resp_body = test_client.get(
-        '/custom/204', http_conn=http_connection
+        '/custom/204', http_conn=http_connection,
     )
     actual_status = int(status_line[:3])
     assert actual_status == 204
@@ -711,7 +865,7 @@ def test_No_Message_Body(test_client):
 
     # Make a 304 request on the same connection.
     status_line, actual_headers, actual_resp_body = test_client.get(
-        '/custom/304', http_conn=http_connection
+        '/custom/304', http_conn=http_connection,
     )
     actual_status = int(status_line[:3])
     assert actual_status == 304
@@ -721,11 +875,13 @@ def test_No_Message_Body(test_client):
 
 
 @pytest.mark.xfail(
-    reason='Server does not correctly read trailers/ending of the previous '
-           'HTTP request, thus the second request fails as the server tries '
-           r"to parse b'Content-Type: application/json\r\n' as a "
-           'Request-Line. This results in HTTP status code 400, instead of 413'
-           'Ref: https://github.com/cherrypy/cheroot/issues/69'
+    reason=unwrap(
+        trim("""
+        Headers from earlier request leak into the request
+        line for a subsequent request, resulting in 400
+        instead of 413. See cherrypy/cheroot#69 for details.
+        """),
+    ),
 )
 def test_Chunked_Encoding(test_client):
     """Test HTTP uploads with chunked transfer-encoding."""
@@ -758,7 +914,7 @@ def test_Chunked_Encoding(test_client):
 
     # Try a chunked request that exceeds server.max_request_body_size.
     # Note that the delimiters and trailer are included.
-    body = b'3e3\r\n' + (b'x' * 995) + b'\r\n0\r\n\r\n'
+    body = b'\r\n'.join((b'3e3', b'x' * 995, b'0', b'', b''))
     conn.putrequest('POST', '/upload', skip_host=True)
     conn.putheader('Host', conn.host)
     conn.putheader('Transfer-Encoding', 'chunked')
@@ -800,18 +956,37 @@ def test_Content_Length_in(test_client):
     conn.close()
 
 
-@pytest.mark.parametrize(
-    'uri,expected_resp_status,expected_resp_body',
-    (
-        ('/wrong_cl_buffered', 500,
-         (b'The requested resource returned more bytes than the '
-          b'declared Content-Length.')),
-        ('/wrong_cl_unbuffered', 200, b'I too'),
+def test_Content_Length_not_int(test_client):
+    """Test that malicious Content-Length header returns 400."""
+    status_line, actual_headers, actual_resp_body = test_client.post(
+        '/upload',
+        headers=[
+            ('Content-Type', 'text/plain'),
+            ('Content-Length', 'not-an-integer'),
+        ],
     )
+    actual_status = int(status_line[:3])
+
+    assert actual_status == 400
+    assert actual_resp_body == b'Malformed Content-Length Header.'
+
+
+@pytest.mark.parametrize(
+    ('uri', 'expected_resp_status', 'expected_resp_body'),
+    (
+        (
+            '/wrong_cl_buffered', 500,
+            (
+                b'The requested resource returned more bytes than the '
+                b'declared Content-Length.'
+            ),
+        ),
+        ('/wrong_cl_unbuffered', 200, b'I too'),
+    ),
 )
 def test_Content_Length_out(
     test_client,
-    uri, expected_resp_status, expected_resp_body
+    uri, expected_resp_status, expected_resp_body,
 ):
     """Test response with Content-Length less than the response body.
 
@@ -831,10 +1006,17 @@ def test_Content_Length_out(
 
     conn.close()
 
+    # the server logs the exception that we had verified from the
+    # client perspective. Tell the error_log verification that
+    # it can ignore that message.
+    test_client.server_instance.error_log.ignored_msgs.append(
+        "ValueError('Response body exceeds the declared Content-Length.')",
+    )
+
 
 @pytest.mark.xfail(
     reason='Sometimes this test fails due to low timeout. '
-           'Ref: https://github.com/cherrypy/cherrypy/issues/598'
+           'Ref: https://github.com/cherrypy/cherrypy/issues/598',
 )
 def test_598(test_client):
     """Test serving large file with a read timeout in place."""
@@ -842,7 +1024,7 @@ def test_598(test_client):
     conn = test_client.get_connection()
     remote_data_conn = urllib.request.urlopen(
         '%s://%s:%s/one_megabyte_of_a'
-        % ('http', conn.host, conn.port)
+        % ('http', conn.host, conn.port),
     )
     buf = remote_data_conn.read(512)
     time.sleep(timeout * 0.6)
@@ -865,18 +1047,103 @@ def test_598(test_client):
     (
         b'\n\n',
         b'\r\n\n',
-    )
+    ),
 )
 def test_No_CRLF(test_client, invalid_terminator):
     """Test HTTP queries with no valid CRLF terminators."""
     # Initialize a persistent HTTP connection
     conn = test_client.get_connection()
 
-    # (b'%s' % b'') is not supported in Python 3.4, so just use +
-    conn.send(b'GET /hello HTTP/1.1' + invalid_terminator)
+    # (b'%s' % b'') is not supported in Python 3.4, so just use bytes.join()
+    conn.send(b''.join((b'GET /hello HTTP/1.1', invalid_terminator)))
     response = conn.response_class(conn.sock, method='GET')
     response.begin()
     actual_resp_body = response.read()
     expected_resp_body = b'HTTP requires CRLF terminators'
     assert actual_resp_body == expected_resp_body
     conn.close()
+
+
+class FaultySelect:
+    """Mock class to insert errors in the selector.select method."""
+
+    def __init__(self, original_select):
+        """Initilize helper class to wrap the selector.select method."""
+        self.original_select = original_select
+        self.request_served = False
+        self.os_error_triggered = False
+
+    def __call__(self, timeout):
+        """Intercept the calls to selector.select."""
+        if self.request_served:
+            self.os_error_triggered = True
+            raise OSError('Error while selecting the client socket.')
+
+        return self.original_select(timeout)
+
+
+class FaultyGetMap:
+    """Mock class to insert errors in the selector.get_map method."""
+
+    def __init__(self, original_get_map):
+        """Initilize helper class to wrap the selector.get_map method."""
+        self.original_get_map = original_get_map
+        self.sabotage_conn = False
+        self.socket_closed = False
+
+    def __call__(self):
+        """Intercept the calls to selector.get_map."""
+        sabotage_targets = (
+            conn for _, (*_, conn) in self.original_get_map().items()
+            if isinstance(conn, cheroot.server.HTTPConnection)
+        ) if self.sabotage_conn else ()
+
+        for conn in sabotage_targets:
+            # close the socket to cause OSError
+            conn.close()
+            self.socket_closed = True
+
+        return self.original_get_map()
+
+
+def test_invalid_selected_connection(test_client, monkeypatch):
+    """Test the error handling segment of HTTP connection selection.
+
+    See :py:meth:`cheroot.connections.ConnectionManager.get_conn`.
+    """
+    # patch the select method
+    faux_select = FaultySelect(
+        test_client.server_instance._connections._selector.select,
+    )
+    monkeypatch.setattr(
+        test_client.server_instance._connections._selector,
+        'select',
+        faux_select,
+    )
+
+    # patch the get_map method
+    faux_get_map = FaultyGetMap(
+        test_client.server_instance._connections._selector.get_map,
+    )
+
+    monkeypatch.setattr(
+        test_client.server_instance._connections._selector,
+        'get_map',
+        faux_get_map,
+    )
+
+    # request a page with connection keep-alive to make sure
+    # we'll have a connection to be modified.
+    resp_status, resp_headers, resp_body = test_client.request(
+        '/page1', headers=[('Connection', 'Keep-Alive')],
+    )
+
+    assert resp_status == '200 OK'
+    # trigger the internal errors
+    faux_get_map.sabotage_conn = faux_select.request_served = True
+    # give time to make sure the error gets handled
+    time.sleep(0.2)
+    assert faux_select.os_error_triggered
+    assert faux_get_map.socket_closed
+    # any error in the error handling should be catched by the
+    # teardown verification for the error_log
