@@ -8,7 +8,7 @@ from cheroot.wsgi import PathInfoDispatcher
 
 
 def wsgi_invoke(app, environ):
-    """Serve 1 requeset from a WSGI application."""
+    """Serve 1 request from a WSGI application."""
     response = {}
 
     def start_response(status, headers):
@@ -18,19 +18,21 @@ def wsgi_invoke(app, environ):
         })
 
     response['body'] = b''.join(
-        app(environ, start_response)
+        app(environ, start_response),
     )
 
     return response
 
 
 def test_dispatch_no_script_name():
-    """Despatch despite lack of SCRIPT_NAME in environ."""
+    """Dispatch despite lack of ``SCRIPT_NAME`` in environ."""
     # Bare bones WSGI hello world app (from PEP 333).
     def app(environ, start_response):
-        start_response('200 OK', [
-            ('Content-Type', 'text/plain; charset=utf-8'),
-        ])
+        start_response(
+            '200 OK', [
+                ('Content-Type', 'text/plain; charset=utf-8'),
+            ],
+        )
         return [u'Hello, world!'.encode('utf-8')]
 
     # Build a dispatch table.
@@ -39,9 +41,11 @@ def test_dispatch_no_script_name():
     ])
 
     # Dispatch a request without `SCRIPT_NAME`.
-    response = wsgi_invoke(d, {
-        'PATH_INFO': '/foo',
-    })
+    response = wsgi_invoke(
+        d, {
+            'PATH_INFO': '/foo',
+        },
+    )
     assert response == {
         'status': '200 OK',
         'headers': [
